import java.util.LinkedList;

/**
 * Une classe représentant un point en 2D à coordonnées entières p=(lw, rw)
 * Utile pour dessiner les arbres binaires (rouges et noirs)
 */
abstract class IntegerPoint2D {

  /** x and y coordinates */
  int lw, rw;

}

/**
 * A class representing a (node of a) Red-Black binary search tree
 */
class RBT extends IntegerPoint2D {
  final static boolean RED = true;
  final static boolean BLACK = false;

  boolean color;
  RBT left, right;
  String element;

  /**
   * Construit un noeud d'un arbre binaire Rouge-Noir
   * 
   * @param color   couleur du noeud
   * @param left    fils gauche
   * @param element élément stoqué dans le noeud  
   * @param right   fils droit
   */
  RBT(boolean color, RBT left, String element, RBT right) {
    this.color = color;
    this.left = left;
    this.element = element;
    this.right = right;
  }

  // Question 2.1
  
  /**
   * Vérifie si l'arbre en entrée satisfait propriété (2) de la définition des arbres rouges et noires.
   * 
   * @param t       noeud racine d'un arbre rouge noir
   * @return true   si les noeuds rouges satisfont la condition (2) de la définition
   */
  static boolean isRedValid(RBT t) {
    // throw new Error("Méthode boolean isRedValid(RBT t) à compléter (Question 2.1)");
    if (t == null)
      return true;
    if (t.color == RED) {
      if (t.left != null && t.left.color == RED)
        return false;
      if (t.right != null && t.right.color == RED)
        return false;
    }
    return isRedValid(t.left) && isRedValid(t.right);
  }

  // Question 2.2

  /**
   * Vérifie la validité de la condition (3) de la définition des arbres rouges et noirs: <br>
   * pour tout noeud 'v' de l'arbre, les chemins qui de vont de 'v' jusqu'aux feuilles contiennent tous
   * le même nombre de sommets noirs.
   * 
   * @param t  noeud racine d'un arbre binaire de recherche
   * @return true   si pour tout noeud 'v' de l'arbre, les chemins qui de vont de 'v' jusqu'aux feuilles contiennent tous
   * le même nombre de sommets noirs.
   */
  static boolean isBlackValid(RBT t) {
    // throw new Error("Méthode boolean isBlackValid(RBT t) à compléter (Question 2.2)");
    return isBlackValid(t, blackHeight(t));
  }

  static boolean isBlackValid(RBT t, int h) {
    if (t == null)
      return h == 0;
    if (t.color == BLACK)
      h--;
    return isBlackValid(t.left, h) && isBlackValid(t.right, h);
  }

  /**
   * Calcule le nombre de sommets noirs le long du chemin le plus à gauche (partant du noeud t)
   * et jusqu'aux feuilles.
   * 
   * @param t   noeud n'un arbre rouge et noire
   * @return    le nombre de sommets noirs le long du chemin le plus à gauche (de t jusqu'aux feuilles)
   */
  static int blackHeight(RBT t) {
    int h = 0;
    while (t != null) {
      if (t.color == BLACK)
        h++;
      t = t.left;
    }
    return h;
  }

  // Question 2.3

  /**
   * Vérifie qu'il s'agit d'un arbre rouge et noire valide
   * 
   * @param t  noeud racine d'un arbre binaire de recherche
   * @return true   si l'arbre est valide (les 3 conditions de la définition sont respectées)
   */
  static boolean isValid(RBT root) {
    // throw new Error("Méthode boolean isValid(RBT t) à compléter (Question 2.3)");
    if(root==null) return true; // empty tree
    if(root.color==RED) return false;

    return isBlackValid(root) && isRedValid(root);
  }

  // Question 2.4

  /**
   * Calcule et renvoie l'hauteur d'un arbre rouge et noire contenant n éléments. <br>
   * <br>
   * L'hauteur d'un tel arbre est donnée par l'entier non négatif h qui satisfait: <br>
   * <br>
   * <tt>
   * 2^h-1 <= n < 2^(h+1)-1
   * </tt>
   * 
   * @param n nombre d'éléments à stoquer dans l'arbre (number of elements stored
   *          in the tree)
   * @return l'hauteur de l'arbre
   */
  static int estimateBlackHeight(int n) {
    assert n >= 0;
    // throw new Error("Méthode int estimateBlackHeight(int n) à compléter (Question 2.4)");
    int h = 0, p = 1; // p=2^h
    while (n >= p - 1) {
      h++;
      p *= 2;
    }
    return h - 1;
  }

  // Question 2.5

  // construire un arbre rouge et noir à partir d'une liste *triée*
  // principe : tous les noeuds noirs, sauf sur la dernière ligne (lorsqu'elle est
  // incomplète)

  /**
   * Consruit un arbre Rouge et   Noir à partir d'une liste triée
   */
  static RBT ofList(LinkedList<String> l) {
    // throw new Error("Méthode RBT ofList(LinkedList<String> l) à compléter (Question 2.5)");
    int h = estimateBlackHeight(l.size());
    return ofList(l, h, l.size());
  }

  /**
   * Construct a Red-Black tree from an ordered input list
   * 
   * @param l liste triée
   */
  static RBT ofList(LinkedList<String> l, int h, int n) {
    assert (1 << h) - 1 <= n && n < (1 << (h + 1));
    if (h == 0) {
      if (n == 0)
        return null;
      return new RBT(RED, null, l.removeFirst(), null);
    }
    int n1 = (n - 1) / 2;
    RBT left = ofList(l, h - 1, n1);
    String e = l.removeFirst();
    RBT right = ofList(l, h - 1, n - n1 - 1);
    return new RBT(BLACK, left, e, right);
  }

  /**
   * Affiche les noeuds de l'arbre binaire (ordre infixe)
   * 
   * @param root  racine d'un arbre binaire
   */
  public static void print(RBT root) {
    if (root != null) {
      print(root.left);
      System.out.println(root.element);
      print(root.right);
    }
  }

  /**
   * Renvoie le nombre de noeuds d'un arbre binaire rouge et noire
   */
  static int size(RBT node) {
    if (node == null)
      return 0;
    else
      return size(node.left) + 1 + size(node.right);
  }

}
